:- module(curryast, [split_rule/3, get_params/2, get_function/2,
	             param_at/3, make_name/2, map_tree/3, map_rules/1,
		     get_variables/2, check_tree/1]).

% We use high-order prolog.
:- use_package(hiord).

%%% Real includes.
:- use_module(library(terms), [atom_concat/2] ).
:- use_module(library(lists), [nth/3] ).

:- use_module(symtab,         [get_real_rules/2, update_rules/2] ).

%% In the future would be nice to export those from here, or maybe
%% providing enought predicates here such that no module needs to know the operators.
:- include(ast_ops).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Structural predicates.
%% Should be used when posible instead including ast_ops.

split_rule(Rule, Head, Body) :- 
	Rule = Head := Body.

get_params(_@Params, Params) :- !.

%% TODO: Maybe we should handle constants.
get_function(F@_, F) :- !.
get_function(F, F) :- !.

param_at(FunApp, [], FunApp).
param_at(_@PList, [Pos|Poss], Exp) :-
	nth(Pos, PList, TExp),
	param_at(TExp, Poss, Exp).

%% Gets all the Expr variables (param ATM).
get_variables(Expr, Vars) :-
	map_tree(Expr, get_var(Vars), _),
	length(Vars, _), % Hacky, hacky.
	!.

%% For use with map_tree.
get_var(param(X), Dicc, const(_,_)) :-
	member(param(X), Dicc).


%% What a hack!
make_name(const(_, C), C).
make_name(qid('Prelude', '[]'), []).
make_name(qid('Prelude', ':'),  .).	
make_name(qid(Module, Name), SName) :-
	atom_concat([Module, '_', Name], SName).
%% Removed by now.
%make_name(qid('Prelude', ','),  ',').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Map over AST first attempt.
%% This needs some care because our AST is not an algebraic data type.

%% HOW TO USE??
%% You must supply a predicate that fails without doing anything if you 
%% don't want to operate on these term, or succeeds giving the transformed
%% term.

:- meta_predicate map_tree(?, pred(2), ?).

%% We try to match the replace, once matched this continues rewriting.
map_tree(E, Func, NE) :-
	Func(E, E1),  % This fails depending on the specified Func.
	!,
	(
	    E == E1 ->
	    E1 = NE   % Or maybe fail, so we continue with other transformations.
	;
	    map_tree(E1, Func, NE)
	).

%% If we have no match, recurse structuraly over the AST.
map_tree(const(A, B), _F, const(A, B)) :-
	!.
map_tree(freevar(A), _F, freevar(A)) :-
	!.
map_tree(param(A), _F, param(A)) :-
	!.
map_tree(qid(M, N), _F, qid(M, N)) :-
	!.
map_tree(gunref(X), _F, gunref(X)) :-
	!.
map_tree(E@EL, Func, NE@NEL) :-
	map_tree(E, Func, NE),
	!,
	map_tree_list(EL, Func, NEL).

map_tree(lambda(Pattern, Expr), Func, lambda(NPattern, NExpr)) :-
	map_tree(Pattern, Func, NPattern),
	!,
	map_tree(Expr, Func, NExpr).

map_tree(ldef(LDefs, Expr), Func, ldef(LDefs2, NExpr)) :-
	map_tree_list(LDefs, Func, LDefs2),
	!,
	map_tree(Expr, Func, NExpr).

%% For ldefs
map_tree(pattern(Pat, Expr), F, pattern(Pat1, Expr1)) :-
	map_tree(Pat, F, Pat1),
	map_tree(Expr, F, Expr1).

map_tree(Head := Body, F, Head1 := Body1) :-
	map_tree(Head, F, Head1),
	map_tree(Body, F, Body1).

%% Rules mapping.

:- meta_predicate map_tree_list(?, pred(2), ?).

map_tree_list([], _F, []) :- 
	!.

map_tree_list([E|EL], F, [NE|NEL]) :-
	map_tree(E, F, NE),
	!,
	map_tree_list(EL, F, NEL).

:- meta_predicate map_rules(pred(2)).

map_rules(Function) :-
	get_real_rules(Name, Rules),
	map_rules_list(Rules, Function, NRules),
	update_rules(Name, NRules),
	fail.
map_rules(_F).

:- meta_predicate map_rules_list(?, pred(2), ?)
.
map_rules_list([], _F, []) :-
	!. %% This is here because our loop is side-effect and fail.
map_rules_list([R|RL], Function, [NR|NRL]) :-
	split_rule(R,  Head, Body),
	map_tree(Head, Function, NHead),
	map_tree(Body, Function, NBody),
	split_rule(NR, NHead, NBody),
	!,
	map_rules_list(RL, Function, NRL).

%% check_tree(AST), takes a Curry AST and verifies that is correct.

check_tree(builtin(_)).

check_tree(Head := Body) :-
	check_head(Head),
	check_expr(Body).

check_head(qid(_Mod, _Name)@Pattern) :-
	check_pattern(Pattern).

check_pattern(P) :-
	list(P),
	check_expr_list(P).

%check_expr(unref(_)) :-
%	write(warning).

check_expr(const(_, _)).
check_expr(freevar(_)).
check_expr(param(_)).
check_expr(qid(_,_)).

check_expr(lambda(Pattern, Expr)) :-
	check_expr(Pattern),
	!,
	check_expr(Expr).

check_expr(ldef(_LDefs, Expr)) :-
%	check_defs(...
	check_expr(Expr).

check_expr(Expr@ExprList) :-
	check_expr(Expr),
	check_expr_list(ExprList).

check_expr_list([]).
check_expr_list([E|Es]) :-
	check_expr(E),
	!,
	check_expr_list(Es).

check_func(qid(_Mod, _Name)).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

