:- module(lexer, [scanTokensExpr/2, scanFile/2, display_tokens/1], _).

:- use_module(library(lists),         [length/2, sublist/2, append/3, insert_last/3] ).
:- use_module(library(dynamic),       [assert/1, retract/1, retractall/1] ).
:- use_module(library(file_utils),    [file_to_string/2]).
:- use_module(library(iso_byte_char), [atom_chars/2,char_code/2]).

:- use_package(hiord).

% TODO:
% ops can be `id`
% only support num constants, should support string and char constants,
% easy as they start by ' and "

% egallego@babel: 02 Nov 2004, 17:09
% 
% Lexer module. It exports two predicates, scanTokens and scanFile 
% obtaining a list of tokens.
% 
% If you want to use old interface it is in lexer_compat module.

% Wrapper to scan from a file.
scanFile(File, Tokens) :-
	file_to_string(File, S),
	scanTokensModule(S, Tokens).

%% scanTokensModule se encarga de tokenizar un módulo, esto tiene el añadido 
%% principal de que se puede necesitar introducir a fake module Main where 
%% if we have no module present, or in other case we must start a col 0.

scanTokensModule(String, ListaTokens) :-
	init_pos,
	init_layout,
	scanPrelude(String, PreludeT),
	scanTokensList(String, TokensScan),
	append(PreludeT, TokensScan, ListaTokens).

scanPrelude(String, []) :-
	get_real_token(String, _, T),
	T = (lex_reserved(module), _),
	!.

scanPrelude(_S, T) :-
	T = [(lex_reserved(module), pos(1,0)),
	     (lex_conid('Main'),    pos(1,8)),
	     (lex_reserved(where),  pos(1,14)),
	     ('{',pos(3,2))],
	pushLayout(1).
       

% scanTokens(String, ListaTokens) se encarga de tokenizar un string.
% de momento lee de un String, pero puede leer de otro lado.

scanTokensExpr(String, ListaTokens) :-
	init_pos,
	init_layout,
	scanTokensList(String, ListaTokens).


% This is the main lexer function.
scanTokensList([], [T]) :- !,
	set_actual_pos(eof, T). % We return eof here but this is a FIXME.
scanTokensList(Input, [Token|Tokens]) :-
	scanToken(Input, NewInput, Token),
	!,
	scanTokensList(NewInput, Tokens).

% scanToken (Input, NewInput, Token)
%scanToken([],[],[]).
scanToken(Input, NewInput, Token) :-
	filter_blanks(Input, TInput),
	(
	    (
		is_bol
	    ;
		is_new_layout
	    ),
	    \+ fake_input(TInput) ->
	    get_layout_token(TInput, NewInput, Token)
	;
	    get_real_token(TInput, NewInput, Token)
	).

:- use_module(library(strings)).
scanToken(Input, Input, Token) :-
	write_string(Input),
	throw(failed_in_scantoken).

fake_input([]).
fake_input([eof|_]).
fake_input([close_layout|_]).

get_real_token([], NewInput, T) :- 
	(
 	    need_close_layout(NewInput, T) 
 	;
	    
 	    NewInput = [],
	    set_actual_pos(eof, T)
 	),
 	!.

%% Layout closing at eof.
	
get_real_token([close_layout|R], R, T) :-
 	set_actual_pos('}', T).


get_real_token([eof], [], T) :-
 	set_actual_pos(eof, T).

get_real_token(Input, NextInput, Token) :-
	select_viable_token(Input, NewInput, CandToken),
	(
	    token(CandToken, NToken) ->
	    (   
	        layout_props(insert_in), 
		retract(layout_props(insert_in)),
		set_actual_pos(lex_reserved(in), Token),
		NextInput = Input
	    ;
		NToken = lex_reserved(in),
		assert(layout_props(insert_in)),
		popLayout,
		set_actual_pos('}', Token),
		NextInput = NewInput
	    ;
		actual_pos(Line, Col),
		Token = (NToken, pos(Line, Col)),
		update_pos(CandToken),
		tag_special_layout_token(NToken),
		NextInput = NewInput,
		(
		    layout_props(start_layout) ->
		    retract(layout_props(start_layout)),
		    pushLayout(Col)
		;
		    true
		)
	    )
	;
	    actual_pos(Line, Col),
	    throw(lexer_error((pos(Line, Col), CandToken, Input, NewInput)))
	).


% This predicate returns the new a layout token if it's needed or
% a real token otherwise.
get_layout_token(Input, NewInput, Token) :-
	(
	    is_new_layout ->
	    retract(layout_props(new_layout)),
	    assert(layout_props(start_layout)),
	    set_actual_pos('{', Token),
	    Input = NewInput
	;
	    is_bol ->
	    retract(layout_props(bol)),
	    actual_pos(_, CT),
	    layout_pos(CL),
	    (
		CT = CL ->
		set_actual_pos(';', Token),
		Input = NewInput	    
	    ;
		(
		    CT < CL ->
		    popLayout,
		    set_actual_pos('}', Token),
		    assert(layout_props(bol)),
		    Input = NewInput
		;
		    get_real_token(Input, NewInput, Token)
		)
	    )
	).


% select_viable_token takes as much of its input as it
% can and returns it.
% In the future this should use regular expressions.
select_viable_token([],[],[]).

% String and char constants
select_viable_token([C|CH], NewInput, string(CToken)) :-
	is_string_delim(C),
	!,
	get_string(CH, NewInput, CToken).

%% If we have a char constant, maybe we have to tranlate it if
%% is a special literal.
select_viable_token([C|CH], NewInput, char_const(CToken)) :-
	is_char_delim(C),
	!,
	CH = [C1|CR],
	(
	    %% Is a \?, then translate raw char.
	    C1 = 0'\\ ->
	    CR = [C2,0''|NewInput],
	    translate_raw(C2, CToken)
        ;
            C1 = CToken,
	    CR = [0''|NewInput]
        ).

% Identifiers.
select_viable_token([C|CH], NewInput, CToken) :-
	is_alfa(C), 
	!,
	get_while(is_alfanum ,[C|CH], NewInput, CToken).

% Numbers.
select_viable_token([C|CH], NewInput, CToken) :-
	is_digit(C),
	!,
	get_while(is_alfanum, [C|CH], NewInput, CToken).

select_viable_token([CToken|CH], CH, [CToken]) :-
	is_punt(CToken), 
	!.

select_viable_token(Input, NewInput, CToken) :-
	get_while(is_not_alfanum, Input, NewInput, CToken), 
	!.

% A high order get_while.

:- meta_predicate get_while(pred(1), ?, ?, ?).

get_while(Goal, [C|CH], ListRes, [C|RS]) :-
	Goal(C),
	!,
	get_while(Goal, CH, ListRes, RS).

get_while(_, List, List, []).

%
get_string(Input, NewInput, String) :-
	get_string_char(Input, Input2, Char),
	(
	    Char = eos ->
	    String = [],
	    Input2 = NewInput
        ;
	    String = [Char|RestStr],
	    !,
	    get_string(Input2, NewInput, RestStr)
        ).

get_string_char([0'" | NewInput], NewInput, eos) :-
	!.
get_string_char([0'\\, RawChar | NewInput], NewInput, Char) :-
	translate_raw(RawChar, Char).
get_string_char([Char|NewInput], NewInput, Char).

%% This is the raw translation table. For adding a new \c constant
%% update it. 

translate_raw(0'n, 0'\n).
translate_raw(0'\\, 0'\\).
translate_raw(0't, 0'\t).
translate_raw(0'", 0'").
translate_raw(0'r, 0'\r).

% Layout processing.

:- dynamic layout_props/1.

is_bol :-
	layout_props(bol).

is_new_layout :-
	layout_props(new_layout).

init_layout :-
	retractall(layout_props(_)),
	assert(layout_props(layout([]))). %FIXME

layout_pos(C) :-
	layout_props(layout([C|_])),
	!.

layout_pos(0) :-
	!.

need_close_layout([eof], T) :-
	layout_props(layout([_])),
	set_actual_pos('}', T),
	!.

need_close_layout([close_layout|R], T) :-
	popLayout,
	need_close_layout(R, T).
	
	
pushLayout(N) :-
	retract(layout_props(layout(L))),
	assert(layout_props(layout([N|L]))).

popLayout :-
	retract(layout_props(layout([_|L]))),
	assert(layout_props(layout(L))).

tag_special_layout_token(lex_reserved(where)) :-
	assert(layout_props(new_layout)).

tag_special_layout_token(lex_reserved(let)) :-
	assert(layout_props(new_layout)).

tag_special_layout_token(lex_reserved(of)) :-
	assert(layout_props(new_layout)).

tag_special_layout_token(lex_reserved(do)) :-
	assert(layout_props(new_layout)).

tag_special_layout_token(_).


set_actual_pos(Symbol, Token) :-
	actual_pos(L, C),
	Token = (Symbol, pos(L, C)).


% Tipos de tokens que tenemos: 
%
is_token(X) :-
	token(_, X), !.

% Punctuation tokens.

token("(", '(').
token(")", ')').
token("[", '[').
token("]", ']').
token("{", '{').
token("}", '}').
token(",", ',').
token(";", ';').
token("\\", '\\').

%% Constants.
token(char_const(C), const(tid('Prelude','Char'), C)).
token(string(S), const(tid('Prelude', 'String'), S)).
token(X, const(tid('Prelude', 'Int'), N)) :-
	number_codes(N, X).

% Reserved words.
token(Name, lex_reserved(Token)) :-
	atom_codes(Token, Name),
	is_reserved(Token).

% identifier tokens.
token([C|CH], lex_id(N)) :-
	is_lower(C),
	atom_codes(N, [C|CH]).

token([C|CH], lex_conid(N)) :-
	is_upper(C),
	atom_codes(N, [C|CH]).

%%% Else we have some unidentified char list, and we should return it as an identifier.
%% Special case . operator.
token(".", lex_op('f_.')).

token(L, lex_op(Op)) :-
	is_op(L),
	atom_codes(Op, L).

token([L|LL], lex_op(Op)) :-
	L = 0'`,
	insert_last(OpN, 0'`, LL),
	atom_codes(Op, OpN).

%%% As defined in curry prelude minus \
is_op(L) :-
	sublist(L, "~!@#$%^&*+-=<>?./|:").

is_punt(C) :-
	char_code(T, C),
	member(T, ['{','}','[',']','(',')',',',';','\\']).

is_reserved(I) :-
	member(I,[class, data, deriving, do, else, free, hiding,
	          if, import, in, infix, infixl, infixr, instance, let,
		  module, newtype, of, qualified, then, type,
		  where, external, engine]).


% Various utils.
% FIXME.
% In filter blanks we should follow the layout.
filter_blanks([],[]).
filter_blanks([0'-, 0'-|CS], NS) :-
	!,
	inc_pos(2),
	skip_comment(CS, NCS),
	filter_blanks(NCS, NS).

filter_blanks([C|CS], NS) :-
	is_blank(C), 
	skip_blank(C),
	!,
	filter_blanks(CS, NS).

filter_blanks(X, X).

% 
skip_blank(0' ) :-
	inc_pos(1).

skip_blank(0'\n) :-
	new_line.

% Following Haskell Layout.
skip_blank(0'\t) :-
	inc_pos(8).

skip_comment([C|CH], NS) :-
	C \== 0'\n, 
	!,
	inc_pos(1),
	skip_comment(CH, NS).

skip_comment(X, X).

% Position handlers.
% Now we need to follow the position in the parser to track layout.

% actual_pos(line, column)
:- dynamic actual_pos/2.

init_pos :-
	(
	    actual_pos(_,_) ->
	    retract(actual_pos(_,_))
	;
	    true
	),
	assert(actual_pos(1, 1)).

inc_pos(N) :-
	retract(actual_pos(Line, Column)),
	NColumn is Column + N,
	assert(actual_pos(Line, NColumn)).

new_line :-
	retract(actual_pos(Line, _)),
	NLine is Line + 1,
	assert(actual_pos(NLine, 1)),
	(
	    \+layout_props(bol) ->
	    assert(layout_props(bol))
	;
	    true
	).

update_pos(string(S)) :-
	length(S, N),
	inc_pos(N).
update_pos(char_const(_C)) :-
	inc_pos(1).
update_pos(String) :-
	length(String, N),
	inc_pos(N).

% FIXME: Of moment we use \n to mark end of rule, but this should
% be done using layout.

is_string_delim(0'").

is_char_delim(0'').

is_blank(X) :-
	member(X," \n\t").

is_alfanum(CH) :-
	is_alfa(CH).
is_alfanum(CH) :-
	is_digit(CH).
is_alfanum(0'').

is_not_alfanum(CH) :-
	\+ is_alfanum(CH),
	\+ is_blank(CH),
	\+ is_punt(CH),
	\+ CH = 0'",
	\+ CH = 0''.

is_alfa(CH) :-
	is_upper(CH).
is_alfa(CH) :-
	is_lower(CH).
% For `op`
is_alfa(0'`).

is_upper(N):-
	N>=65,
	N=<90.

is_lower(N):-
	(N>=0'a,N=<0'z ; N=0'_),!.
is_lower(N):-
	N>127,!.

is_digit(N):-
	N<58,
	N>47.


% Pretty print function.
display_tokens([]) :- !.

display_tokens([(T, _)|TS]) :-
	display_token(T),
	display(' '),
	!,
	display_tokens(TS).

display_token(lex_id(X)) :-
	!,
	display(X).

display_token(lex_num(X)) :-
	!,
	display(X).

display_token(lex_conid(X)) :-
	!,
	display(X).
display_token(lex_reserved(X)) :-
	!,
	display(X).

display_token(lex_op(X)) :-
	!,
	display(X).

display_token(X) :-
	!,
	display(X).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

