:- module(parser, [parse/1, parse_expr/2]).

%%

:- use_module(lexer_compat, [init_scanner/1, init_scanner_string/1,
	                     token_have/1, token_look/1, token_skip/0,
	                     last_token/1, token_back/0]).
:- use_module(symtab,       [add_rule/2, get_object_property/3,
	                     set_object_property/3, add_constructor/2,
	                     new_type/1] ).
:- use_module(debug,        [debug_write/1, debug_write_string/1] ).
:- use_module(cmdopts,      [cmd_options/2] ).
:- use_module(modules,      [start_module/1, end_module/1,
	                     add_exports/1, load_module/1,
	                     current_mod_name/1] ).
:- use_module(priotree,     [pt_insert/3] ).
:- use_module(scope,        [open_scope/1, close_scope/0,
	                     insert_symbol/2, lookup_symbol_unref/2,
	                     lookup_symbol/2, scope_name/1]).
%:- use_module(curryast,     [get_function/2, get_params/2]).
:- use_module(curryast,     [map_tree/3, split_rule/3]).

%% Ciao libraries.
:- use_module(library(lists),      [append/3, length/2] ).
:- use_module(library(write),      [write/1] ).
:- use_module(library(terms),      [atom_concat/2] ).

%% We use some high-order parsing.
:- use_package(hiord).

%% Operators includes.
:- include(ast_ops).
:- include(type_ops).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Module support.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<module>' :-
	(
	    token_have(lex_reserved(module)),
	    token_have(lex_conid(ModName)),
	    start_module(ModName),
	    '<module_exports>',
	    token_have(lex_reserved(where)),
	    (
		\+ cmd_options('-np', _) ->
		load_module(module('Prelude', [], none))
	    ;
		load_module(module('CoreTypes', [], noimport))
	    ),
	    token_have('{'),
	    '<imports>',
	    open_scope(parser(ModName)),
	    '<block>',
	    token_have('}'),
	    %close_scope, %% We need the scope open for symbol resolution.
	    end_module(ModName),
	    token_have(eof)
	;
	    throw_parser_error('Error in module declaration')
	).	

'<module_exports>' :-
	(
	    token_look('(') ->
	    open_close_scan(Exports, list_scan('<module_export_unit>', ','), '(', ')'),
	    add_exports(Exports)	
        ;
	    add_exports(all)
        ).

%% A module export unit can be: (Prelude pg 25)
%% - function ids.
%% - datatypes with or without constructors.
%% - another modules.

'<module_export_unit>'(id(X)) :-
	token_have(lex_id(X)).

'<module_export_unit>'(type(Name, ListOfCons)) :-
	token_have(lex_conid(Name)),
	open_close_scan(ListOfCons, '<conid>', '(', ')').

%% Import declarations.

'<imports>' :-
	decl_list('<import_block>'),
	!.
'<imports>'. % No imports.

'<import_block>' :-
	token_have(lex_reserved(import)),
	!,
	'<import_decl>'(Imports),
	load_module(Imports).

'<import_decl>'(module(Name, Symbols, Modifier)) :-
	(
	    '<module_import_mod>'(Modifier) ->
	    true
	;
	    Modifier = none
	),
	'<conid>'(Name),
	(
	    token_look(';') ->
	    Symbols = []
	;
	    open_close_scan(Symbols, list_scan('<module_export_unit>', ','), '(', ')')
	).

'<module_import_mod>'(qualified) :-
	token_have(lex_reserved(qualified)).

'<module_import_mod>'(hiding) :-	
	token_have(lex_reserved(hiding)).

'<module_import_mod>'(engine) :-
	token_have(lex_reserved(engine)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A block is a list of block_unit.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<block>' :-
	decl_list('<block_unit>').

%%  block_unit ::=  
%%       data type_declaration
%%       fixity fixity_arguments
%%       function_name '::' type_expr
%%       function_name equat

'<block_unit>' :-
	token_have(lex_reserved(data)),
	!,
	'<type_declaration>'(Type, Constructors),
	!,
	record_type(Type, Constructors).

% We need typer support for this.
% '<block_unit>' :-
% 	token_have(lex_reserved(type)),
% 	!,
% 	'<type_alias>'(Name, Type),
% 	make_type_alias(Name, Type).

'<block_unit>' :-
	token_look(lex_reserved(FixityKeyword)),
	member(FixityKeyword, [infix,infixl,infixr]),
	!,
	token_skip,
	'<fixity_argument>'(Nat, Ops),
	record_fixity(FixityKeyword, Nat, Ops).

'<block_unit>':-
	(
	    '<function_name>'(FunctionName)
	;
	    NT = yes
	),
	(
	    token_have(lex_op('::')) ->
	    '<type_expr>'(TypeExpr),
	    record_fun_type(FunctionName, TypeExpr)
	;
	    %% Care here with scopes.
	    token_have(lex_reserved(external)),
	    FunctionName = qid(_, Name),
	    insert_symbol(qid(Name), FunctionName),
	    add_rule(FunctionName, builtin(external))
	;
	    (
		var(NT) ->
		token_back
	    ;
		true
	    ),
	    '<equat>'(Equat),
	    %% FIXME: This is an ugly hack.
	    Equat = (Fname@_) := _,
	    record_equat(Fname, Equat)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Symbol names, must add qualified names.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This is ugly, should cleanup.
'<function_name>'(OpName):-
	token_have('('),
	(
	    token_look(lex_op(OpName1)) ->
	    token_skip,
	    token_have(')'),
	    put_scope_name(OpName1, OpName)
	;
	    token_back,
	    !,
	    fail
	).

'<function_name>'(FunctionName):-
	token_have(lex_id(FunctionName1)),
	put_scope_name(FunctionName1, FunctionName).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reglas de declaracion de tipos:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% type_expr ::=
%      type_expr_unit 
%      type_expr_unit '->') type_expr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<type_expr>'(TypeExpr):-
	'<type_application>'(Tu),
	(
	    token_have(lex_op(->)) ->
	    '<type_expr>'(Te),
	    TypeExpr = (Tu --> Te)
	;
	    TypeExpr = Tu
	).

%% Type application
'<type_application>'(TypeExpr) :-
	'<conid>'(C),
	lookup_symbol_unref(tid(C), QCon),
	!,
	(
	    seq_scan(TypeList, '<simple_type_expr>') ->
	    TypeExpr = QCon<@TypeList   
	;
	    TypeList = [] -> %% FIXME: Verificar la validez de esto.
	    QCon = TypeExpr
	).

'<type_application>'(TypeExpr) :-
	'<simple_type_expr>'(TypeExpr).

% type_expr_unit representa la declaracin de un tipo.
% type_expr_unit ::=
%      id                                  : typevars
%      ()                                  : Unit type.
%      '(' (type_expr) [','type_expr]  ')' : Tuplas
%      '[' type_expr  ']'                  : Listas
%      Bool | String | Char | Int | Float  : Tipos bsicos, deberan ir en el preludio.

'<simple_type_expr>'(Cons) :-
	token_have(lex_conid(C)), 
	!,
	lookup_symbol_unref(tid(C), Cons).

'<simple_type_expr>'(tvar(I)) :- % Regla 2
	token_have(lex_id(I)),
	!.

'<simple_type_expr>'(TypeExpr):-
	token_have('('),
	!,
	(
	    token_have(')') ->
	    TypeExpr = tid('Prelude', ()<@[])
	;
	    (
		'<type_expr_list>'(TypeListAux),
		(
		    length(TypeListAux, 1),
		    TypeListAux = [A] ->
		    TypeExpr=A
		;
		    %type_list_to_tuple(TypeListAux, TypeExprT),
		    TypeExpr = tid('Prelude', tuple)<@TypeListAux
		),
		token_mustbe(')',') expected')
	    )
	).

'<simple_type_expr>'(tid('Prelude', [])<@[TypeExpr]) :-
	open_close_scan(TypeExpr, '<type_expr>', '[', ']').


'<type_expr_list>'([T1|Ts]) :- % <type_expr_list>::=
	'<type_expr>'(T1),    % <type_expr> | <type_expr>,<type_expr_list>
	(
	    token_have(',') ->
	    '<type_expr_list>'(Ts)
	;
	    Ts = []
	).

%% type_declaration ::=
%%       CONID type_var_ids '=' constr_declaration ['|' constr_declaration]
%%

'<type_declaration>'(Type, ConsList):-
	token_mustbe(lex_conid(Cname), 'A constructor name was expected'),
	'<type_var_ids>'(Ids),
	put_scope_type_name(Cname, QType),
	insert_symbol(tid(Cname), QType),
	Type = QType<@Ids,
	token_mustbe(lex_op('='),      '= expected after type variables'),
	'<constr_declaration>'(ConstrDeclaration),
	(
	    token_have(lex_op('|')) -> 
	    '<constr_declarations>'(ConstrDecs),
	    ConsList = [ConstrDeclaration|ConstrDecs]
	;
	    ConsList = [ConstrDeclaration]
	).
        %% Types cannot be freed now.
	%%	free_type(Type, FType, _).

%% Type vars:
%% type_var_ids ::=
%%      { ID }

'<type_var_ids>'(List):-!,
	(
	    token_have(lex_id(Id)) ->
	    (
		'<type_var_ids>'(Ids),
		List = [tvar(Id)|Ids]
	    )
	;
	    List = []
	).  

'<constr_declarations>'(ConstDeclarations):-
	'<constr_declaration>'(ConstDec),
	(
	    token_have(lex_op('|')) ->
	    (
		'<constr_declarations>'(ConstDecs),
		ConstDeclarations = [ConstDec|ConstDecs]
	    )
	;
	    (
		ConstDeclarations = [ConstDec]
	    )
	).

% Declaracin de los constructores del tipo de datos.
%
% constr_declaration ::= 
%       CONID type_exprs
%      

'<constr_declaration>'(ConstrDeclaration):-
	token_mustbe(lex_conid(Cname),'A constructor name was expected'),
	put_scope_name(Cname, QCname),
	insert_symbol(qid(Cname), QCname),
	(
	    seq_scan(TypeExprs, '<simple_type_expr>') ->
	    true
	;
	    TypeExprs = []
	),
	ConstrDeclaration = (QCname, TypeExprs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Expressions: %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% expr ::=
%     '\' pattern '->' expr
%     IF expr THEN expr ELSE expr
%     let {local_defs} in expr
%     F Args
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<expr>'(lambda(LName@P, Expr)) :-
	token_have('\\'),
 	!,
 	seq_scan(P, '<simple_pattern>'),
  	token_mustbe(lex_op(->),'-> expected in lambda expresion'),
  	open_scope(parser('_lambda_e_')),
  	insert_scope_args_list(P),
  	'<expr>'(Expr),
	get_new_lambda_name(LName),
  	close_scope.


'<expr>'(qid('Prelude', if_then_else)@[If,Then,Else]) :-
	token_have(lex_reserved(if)),
	!,
	'<expr>'(If),
	token_mustbe(lex_reserved(then),'then keyword expected'),
	'<expr>'(Then),
	token_mustbe(lex_reserved(else),'else keyword expected'),
	'<expr>'(Else).

'<expr>'(ldef(SDefs, InnerExpr)) :-
 	token_have(lex_reserved(let)),
 	!,
	open_scope(parser('_let_scope_')),
 	open_close_scan(LDefs, list_scan('<local_def>', ';'), '{', '}'),
	%% Make new defined symbols available to the inner expression.
	insert_scope_ldefs(LDefs),
	resolve_local_refs(LDefs, SDefs),
	token_have(lex_reserved(in)),
	'<expr>'(InnerExpr),
	close_scope.
	
'<expr>'(Expr):-
	'<func_expr>'(E1),
	'<infix_exp>'(E1, Expr).

%% Insert the new names brougth by local definitions.
insert_scope_ldefs([]).
insert_scope_ldefs([Ldef|LR]) :-
	insert_scope_ldef(Ldef),
	!,
	insert_scope_ldefs(LR).

insert_scope_ldef(freevars(L)) :-
	insert_scope_args_list(L).

insert_scope_ldef(pattern(Pat, _Expr)) :-
	insert_scope_args(Pat).

%% The function ldef gets is name inserted when defined.
insert_scope_ldef(_Head := _Body).

%% This resolves local scope unrefs in lets. It must be done here,
%% because if later, we only keep the global scope.
resolve_local_refs([], []).
resolve_local_refs([freevars(_)|L], LL) :-
	!,
	resolve_local_refs(L, LL).
resolve_local_refs([L|LS], [R|RS]) :-
	solve_ldef_refs(L, R),
	!,
	resolve_local_refs(LS, RS).

solve_ldef_refs(pattern(Pat, E1), pattern(Pat, E2)) :-
	map_tree(E1, solve_refs_soft, E2).

solve_ldef_refs(FDef, NFDef) :-
	split_rule(FDef, Head, Body),
	map_tree(Body, solve_refs_soft, NBody),
	split_rule(NFDef, Head, NBody).

%% We don't error if a symbol it is not defined, as can be defined ahead.
%% The global pass will catch this.
solve_refs_soft(unref(X), E) :-
	(
	    lookup_symbol(X, E)
	;
	    E = unref(X)
	).

%% Resolve all possible symbols and construct a ldef expression.
solve_where(Expr, Bindings, ldef(LDefs, NExpr)) :-
	insert_scope_ldefs(Bindings),
	map_tree(Expr, solve_refs_soft, NExpr),
	resolve_local_refs(Bindings, LDefs),
	close_scope. %% Matching the open_scope in '<where>'.

%% Let and where declarations
'<local_def>'(LDef) :-
	'<id>'(_Id),
	!,
	(
	    (
		token_look(',')
	    ;
		token_look(lex_reserved(free)) 
	    ) ->
	    token_back,
	    '<free_vars>'(LDef)
	;
	    token_back,
	    '<equat>'(LDef)
	).

'<local_def>'(pattern(Pattern, RExpr)) :-
	!,
	'<pattern>'(Pat),
	get_pattern_name(PName),
	Pattern = PName@[Pat],
	token_have(lex_op('=')),
	'<expr>'(Expr),
	(
	    '<where_exp>'(Bindings),
	    solve_where(Expr, Bindings, RExpr)
	;
	    Expr = RExpr
	).

'<free_vars>'(freevars(L)) :-
	list_scan(L, '<freevar>', ','),
	token_have(lex_reserved(free)).
	
%% Where definition.
'<where_exp>'(Bindings) :-
	token_have(lex_reserved(where)),
	open_scope(parser('_where_scope_')),
	!,
	open_close_scan(Bindings, list_scan('<local_def>', ';'), '{', '}').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Normal infix expressions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<infix_exp>'(OExp, Exp) :-
	(
	    ( %% Bad bad operators.
		token_look(lex_op(Op)),
		\+ forbidden_in_exp(Op),
		token_skip  
	    ) ->
	    (
		'<func_expr>'(FExp) ->
		lookup_symbol_unref(qid(Op), Qop),
		pt_insert(OExp, (Qop, FExp), NExp),
		'<infix_exp>'(NExp, Exp)	    
	    ;
		token_back,
		OExp = Exp
	    )
	;
	    OExp = Exp
	).

forbidden_in_exp(Op) :-
	member(Op, ['=','|']).

%% Applicative expressions.
'<func_expr>'(Expr):-
	seq_scan([Function|Params], '<basic_expr>'),
	Expr = Function@Params.
%%
'<conid>'(X) :-
	token_have(lex_conid(X)).

'<id>'(X) :-
	token_have(lex_id(X)).

'<freevar>'(freevar(X)) :-
	token_have(lex_id(X)).

%% Basic expression support.

'<basic_expr>'(Id@[]) :-
	token_have(lex_id(X)),
	!,
	lookup_symbol_unref(qid(X), Id).

'<basic_expr>'(Cons@[]) :-
	token_have(lex_conid(C)), 
	!,
	lookup_symbol_unref(qid(C), Cons).

'<basic_expr>'(const(T, V)) :-
	token_have(const(T, V)),
	!.

%% Parsing paren expressions is a little tricky, because it can be any
%% of a tuple, the unit (), a operator in functional form or a paren
%% expression. 

'<basic_expr>'(Expr):-
	token_have('('),
	!,
	(
	    token_have(')') -> % This is the unit.
	    Expr = qid('Prelude', '()' )
	;
	    token_have(lex_op(X)),
	    (
		token_have(')') ->
		lookup_symbol_unref(qid(X), Expr)
	    ;
		'<expr>'(PartE), 
		token_have(')'),
		lookup_symbol_unref(qid(X), Op),
		Expr = qid('Prelude', flip)@[Op, PartE]
	    )
	;
	    list_scan(InnerExpr, '<expr>', ','),
	    (
		token_have(lex_op(T)),
		InnerExpr = [_] ->
		lookup_symbol_unref(qid(T), Op),
		Expr = Op@InnerExpr
	    ;
		list_to_tuple(InnerExpr, Expr)
	    ),
	    token_have(')')
	).

%% FIXME, must support list comprehension.
'<basic_expr>'(Expr) :-
	token_have('['),
	!,
	(
	    token_have(']') ->
	    Expr = qid('Prelude', '[]')
	;
	    list_scan(ExprS, '<expr>', ','),
	    desugar_list(ExprS, Expr),
	    token_have(']')
	).

%% Helper predicates.

list_to_tuple([A] , A) :- !.
list_to_tuple([L|LL], qid('Prelude', ',')@[L,B]) :-
	list_to_tuple(LL, B).

type_list_to_tuple([A] , A) :- !.
type_list_to_tuple([L|LL], (L,B)) :-
	type_list_to_tuple(LL, B).

desugar_list([], qid('Prelude', '[]')).
desugar_list([A|B], qid('Prelude', ':')@[A, BS]) :-
	desugar_list(B, BS).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Functions declaration.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<equat>'(Equat):-
	'<funlhs>'(Pattern),
	(
	    token_have(lex_op(=)) -> 
	    '<expr>'(Expr),
	    (
		'<where_exp>'(Bindings) ->
		solve_where(Expr, Bindings, RExpr),
		Equat = Pattern := RExpr
	    ;
		Equat = Pattern := Expr
	    )
	;
	    token_look(lex_op('|')) -> 
	    '<cond_expr>'(CondExpr),
	    Equat = Pattern := (qid('Prelude', left_cond)@[CondExpr])
	;		
	    derror(pipe_expected)
	),
	close_scope.

'<funlhs>'(Pattern) :-
	'<simple_pattern>'(P1),
	%% We look for an operator. When found, we're in a
        %% "a $ b = a b" style declaration.
	(
	    token_look(lex_op(Op)),
	    \+ Op = '=' -> % Infix op decl.
	    token_skip,
	    '<simple_pattern>'(P2),
	    put_scope_name(Op, Qop),
	    Pattern = Qop@[P1, P2],
	    insert_symbol(qid(Op), Qop),
	    open_scope(parser(Op))
	;
	    P1 = param(F), % Normal function decl. 
	    put_scope_name(F, QF) ->
	    (
		seq_scan(PatList, '<simple_pattern>') ->
		Pattern = QF@PatList
	    ;
		Pattern = QF@[]
	    ),
	    insert_symbol(qid(F), QF),
	    open_scope(parser(F))
	;
	    derror(function_decl_error)
	),
	insert_scope_args(Pattern).

%% insert_scope_args inserts all the parameters of a function into the
%% symbol table, so we access them from the function definition. 
%% This predicate doesn't open a new scope.

insert_scope_args(param(X)) :- 
	insert_symbol(qid(X), param(X)).

insert_scope_args(freevar(X)) :- 
	insert_symbol(qid(X), freevar(X)).

insert_scope_args(_Cons@Patt) :-
	insert_scope_args_list(Patt).

insert_scope_args([L]) :-
	insert_scope_args(L).

insert_scope_args((A,B)) :-
	insert_scope_args(A),
	insert_scope_args(B).

insert_scope_args(_).

insert_scope_args_list([]).
insert_scope_args_list([E|EL]) :-
	insert_scope_args(E),
	insert_scope_args_list(EL).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Patterns.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<simple_pattern>'(param(X)) :-
	token_have(lex_id(X)).	

'<simple_pattern>'(Cons) :-
	token_have(lex_conid(X)),
	!,
	lookup_symbol_unref(qid(X), Cons).

%% Fix for constant framework.
'<simple_pattern>'(const(T, V)) :-
	token_have(const(T, V)).

'<simple_pattern>'(Tuple) :-
	token_have('('),
	(
	    token_have(')') ->
	    Tuple = qid('Prelude', '()')
	;
	    list_scan(PatList, '<pattern>', ','),
	    token_have(')'),
	    (
		PatList = [L] ->
		Tuple = L
	    ;
		list_to_tuple(PatList, Tuple)
	    )
	).

'<simple_pattern>'(qid('Prelude', '[]')) :-
	token_have('['),
	(
	    token_have(']') ->
	    true
	;
	    token_back,
	    fail
	).
		
'<simple_pattern>'(PatList) :-
	open_close_scan(PatListS, list_scan('<pattern>', ','), '[', ']'),
	desugar_list(PatListS, PatList).


'<pattern>'(Pattern) :-
	'<head_pattern>'(Head),
	(
	    token_have(lex_op(':')) ->
	    '<pattern>'(Tail),
	    Pattern = qid('Prelude', ':')@[Head, Tail]
	;
	    Pattern = Head
	).

'<head_pattern>'(QConid@Body) :-
	token_have(lex_conid(Conid)),
	lookup_symbol_unref(qid(Conid), QConid),
	% We can have only one constructor.
        (
	    seq_scan(Body, '<pattern>')
	;
	    Body = []
	).

'<head_pattern>'(Pattern) :-
	'<simple_pattern>'(Pattern).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Right hand side constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
'<cond_expr>'(qid('Prelude', ':')@[C, CL]) :-
	token_have(lex_op('|')),
	'<expr>'(CExpr),
	token_mustbe(lex_op(=),'= expected in constraint'),
	'<expr>'(EBody),
	C = qid('Prelude', ',')@[CExpr, EBody],
	(
	    token_look(lex_op('|')) ->
	    '<cond_expr>'(CL)
	;
	    CL = qid('Prelude', '[]')
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Operators declaration.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<fixity_argument>'(Nat, Ops):-
	token_have(const(tid('Prelude','Int'), Nat)),
	'<infix_op_ids>'(Ops).

'<infix_op_ids>'([op(Op)|Ops]):-
	token_mustbe(lex_op(Op),'Malformed infix operator'),
	(
	    token_have(',') -> 
	    '<infix_op_ids>'(Ops) 
	; 
	    Ops = []
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Grammar end.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% High order predicates to parse common patterns.
%% FIXME, we should add error handling here.

:- meta_predicate open_close_scan(?, pred(1), ?, ?).

%% Has use with *Tok free too.
open_close_scan(Res, Goal, OTok, CTok) :-
	token_have(OTok),
	!,
	Goal(Res),
	token_have(CTok).

:- meta_predicate list_scan(?, pred(1), ?).

list_scan([Res|ResL], Goal, Symbol) :-
	Goal(Res),
	(
	    token_have(Symbol) ->
	    list_scan(ResL, Goal, Symbol)
	;
	    ResL = []
	).	

% Sequences (with no commas between elements).
:- meta_predicate seq_scan(?, pred(1)).

seq_scan([Res|ResL], Goal) :-
	Goal(Res),
	(
	    seq_scan(ResL, Goal),
	    !
	;
	    ResL = []
	).

%% We don't use list_scan here as we want different error handling.
:- meta_predicate decl_list(pred(0)).
decl_list(Goal) :-
	token_have(';'),
	!,
	decl_list(Goal).

decl_list(Goal) :-
	Goal,
	!,
	(
	    token_have(';') ->
	    decl_list(Goal)
	;
	    true
	).

%% Utilities for recording things in the symbol table.
record_type(Type, Constructors) :-
	new_type(Type),
	%free_type((Type, Constructors), (FType, FCons), _),
	addconstructors(Type, Constructors).

% Add a list of cons to symbol table.

addconstructors(_Type, []).

addconstructors(Type, [(Cons, TypeOfCons) | R]) :-
	(
	    TypeOfCons = [] ->
	    NewTypeC = Type
	;
	    %% Build the resulting type of the constructor.
    	    append(TypeOfCons, [Type], TypeOfCons1),
	    list2type(TypeOfCons1, NewTypeC)
	),
	add_constructor(Type, Cons),
	set_object_property(Cons, type, NewTypeC),
	set_object_property(Cons, constructs, Type),
	addconstructors(Type, R).

list2type([A], A).
list2type([A|LR], A --> TR) :-
	list2type(LR, TR).

record_fun_type(Name,Type) :-
	(
	    get_object_property(Name, type, _) 	->
	    derror(dup_type(Name))
	;
	    %free_type(Type, FType, _),
	    set_object_property(Name, type, Type)
	).

record_equat(Name, Eq) :-
	check_arity(Name, Eq), 
	add_rule(Name, Eq).
	

%% Esto corresponde al tipador.
%% De momento exigimos que tengan la declaracin de tipos hecha, en otro caso 
%% error.

%% Por cierto el nmero de argumentos tiene que ser fijo, puede que haya que 
%% pensar esto ms.

check_arity(Name, _Def) :-
	(	
	    get_object_property(Name, type, _Type)
	;
	    true
	    %throw(type_error('We have no typer at the moment, sorry', Name))
	).

%% Recording of new operators.
record_fixity(_Fix, _Nat, []).
record_fixity(Fixity, Nat, [op(Op)|Rs]) :-
	put_scope_name(Op, QOp),
	set_object_property(QOp, priority, Nat),
	set_object_property(QOp, assoc,    Fixity),
	record_fixity(Fixity, Nat, Rs).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Main parsing function.

parse(File):-
	
	catch(init_scanner(File), lexer_error(E), do_lexer_error(E)), 
	catch('<module>', parser_error(T, E, R), do_parser_error(T, E, R)).


%% For future shell
parse_expr(String, RExpr) :-
	catch(init_scanner_string(String), lexer_error(E), do_lexer_error(E)),
	'<expr>'(Expr), 
	!,
	(
	    '<where_exp>'(Bindings),
	    solve_where(Expr, Bindings, RExpr)
	;
	    Expr = RExpr
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Error handling.

%% Dummy error function.
derror(Text) :-
	throw_parser_error(Text).

%% Parser recovery. 

%% Resync looks for the token especified and then continues parsing.
do_recovery(resync(Token, Parse)) :-
	resynch(Token),
	call(Parse).

do_recovery(no_recovery) .
%do_recovery(no_recovery) :-
%  	throw(error('Unrecoverable parser error')).

resynch(Token) :-
	(
	    token_look(Token) ->
	    true
	;
	    (
		token_look(eof) ->
		last_token([T|_]),
		throw(parser_error(T, 'Premature End of file', no_recovery))
	    ;
		token_skip,
		resynch(Token)
	    )
	).

%% New error format
%% This needs a big cleanup.

do_parser_error(Token, Error, Recovery) :-
	Token = (T, Pos),
	debug_write('Parser error in: '),
	write_pos(Pos),
	debug_write('\nPossible token: '),
	debug_write(T),
	debug_write('\nError: '),
	debug_write(Error),
	debug_write('\n'),
	catch(do_recovery(Recovery), parser_error(T, E, R), do_parser_error(T, E, R)).
%	catch(do_recovery(Recovery), E, do_parser_error(T, E, R)).
%	do_recovery(Recovery).

write_pos(pos(Line, Column)) :-
	debug_write(line(Line)),
	debug_write(' '),
	debug_write(column(Column)).

throw_parser_error(Error) :-
	last_token([Token|_]),
	throw(parser_error(Token, Error, resync(';', '<block>'))).

do_lexer_error((Pos, CandToken, Input, _NewInput)) :-
	debug_write('Lexer error in: '),
	debug_write(Pos),
	debug_write('\nPossible token: '),
	debug_write(CandToken),
	debug_write('\nInput: '),
	debug_write_string(Input),
	debug_write('\n').	


token_mustbe(Token, ErrCode) :-
	(
	    token_have(Token),
	    !
	; 
	    derror(ErrCode)
	). 

put_scope_name(Name, qid(Module, InnerName)) :-
	findall(N, scope_name(parser(N)), Names),
	!,
	reverse(Names, [Module|RNames]),
	append(RNames, [Name], FullName),
	atom_concat(FullName, InnerName).

put_scope_type_name(Name, tid(Module, InnerName)) :-
	findall(N, scope_name(parser(N)), Names),
	!,
	reverse(Names, [Module|RNames]),
	append(RNames, [Name], FullName),
	atom_concat(FullName, InnerName).

:- dynamic lambda_names/1.

get_new_lambda_name(LName) :-
	retract(lambda_names(N)),
	!,
	N1 is N + 1,
	assert(lambda_names(N1)),
	atom_number(Num, N),
	atom_concat([lambda, Num], Name),
	put_scope_name(Name, LName).
	
get_new_lambda_name(LName) :-
	assert(lambda_names(1)),
	put_scope_name('lambda0', LName),
	!.

:- dynamic pattern_names/1.

get_pattern_name(LName) :-
	retract(pattern_names(N)),
	!,
	N1 is N + 1,
	assert(pattern_names(N1)),
	atom_number(Num, N),
	atom_concat([pattern, Num], Name),
	put_scope_name(Name, LName).
	
get_pattern_name(LName) :-
	assert(pattern_names(1)),
	put_scope_name('pattern0', LName),
	!.


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

