:- module(shell, [start_shell/0, start_shell_debug/0]).

:- use_module(library(terms_vars), [varset/2] ).
:- use_module(library(strings),    [get_line/1]).
:- use_module(library(system),     [time/1]).
:- use_module(library(format),     [format/2]).
:- use_module(library(terms),      [atom_concat/2]).
:- use_module(library(system),     [system/1]).
:- use_module(library(filenames),  [file_name_extension/3]).
:- use_module(library(compiler),   [unload/1]).
:- use_module(library(freeze),     [frozen/2]).

%% For constraint handling. Only in CIAO-1.13
:- use_module(library('attribute_dump/clp_reify')).

%% Expression compiling.
:- use_module(parser,     [parse_expr/2]).
:- use_module(trans_lazy, [tr_expr/2] ).
:- use_module(curryast,   [map_tree/3] ).
:- use_module(unref,      [free_vars/3] ).
:- use_module(localdefs,  [proccess_ldef_global/2, proccess_ldef_pattern/2, proccess_ldef_func/2]).
:- use_module(scope,      [open_scope/1, clean_scope/0, close_scope/0]).
:- use_module(lambda,     [new_lambda/2]).
:- use_module(trans_lazy, [dump/0]).
:- use_module(comp_manager, [rewrite/0]).

%% Module handling and loading.
:- use_module(modules,    [load_module/1, start_module/1, add_exports/1, end_module/1, unload_module/1]).
:- use_module(symtab,     [clean_symtab/0]).
:- use_module(cmdopts,    [cmd_options/2, set_option/2]).

%% Misc.
:- use_module(typer,      [tipar/2]).
:- use_module(prolog2curry, [prolog_to_curry/2, prolog_to_type/2]).

start_shell :-
	reset,
	loop.

start_shell_debug :-
	set_option('-l', ['../lib']),  
	start_shell.  

loop :-
	repeat,
	get_line(Line1),
	(
	    is_command(Line1, Command) ->
	    do_command(Command)
	;
	    append(Line1, [0'\n], Line), %% To be layout compliant.
	    open_scope(parser(shell_expr)), %% 
	    (
		exec_expr(Line) ->
		close_scope
	    ;
		close_scope
	    )
	),
	Command == quit,
	!.

exec_expr(Input) :-
	!,
	translate(Input, Dicc, Prolog),
	pre_variables(Dicc, Prolog, VarInfo),
	!,
	Query = tonf(Prolog, Res),
	(
	    %% load the shell module
	    use_module('Shell'),
	    call(Query),
	    % prolog_to_curry(Res, CRes), % This is crap ATM.
	    write(Res),
	    nl,
	    post_variables(VarInfo),
	    try_more
	;
	    write(no_solutions),
	    nl
	),
	% unload the shell module
	unload('Shell'),
	!,
	fail.

translate(Input, Dicc, Prolog) :-
	start_module('Shell'),
	add_exports(all),
	catch(parse_expr(Input, Expr), E, do_shell_error(E)),
	!,
	(
	    map_tree(Expr, proccess_ldef_func, FExpr),
	    map_tree(FExpr, proccess_ldef_global, PExpr),
	    map_tree(PExpr, proccess_ldef_pattern, LExpr),
	    map_tree(LExpr, new_lambda, TExpr),
	    map_tree(TExpr, free_vars(Dicc), ExprFree),
	    % Generate the shell module
	    end_module('Shell'),
	    gen_shell,
	    tr_expr(ExprFree, Prolog) ->
%	    write(compiled_to(Prolog)),
	    true
	;
	    do_shell_error(error('Failed compilation'))
	).

gen_shell :-
	rewrite,
	tell('Shell.pl'),
	dump,
	portray_clause((:- discontiguous exports/1)),
	portray_clause(exports(_)),
	told.

pre_variables(Dicc, Prolog, vi(Dicc, Binds)) :-
	varset(Prolog, Vs1),
	filter_bindings(Vs1, Dicc, Binds).

post_variables(vi(Dicc, Binds)) :-
	write_bindings(Binds),
	nl,
	(
	    check_constraints(Binds, CProperties) ->
	    (
		CProperties = [] ->
		true
	    ;
		unbind(CProperties, Dicc,CP),
		prettyvars(CP),
		format("Constraint info: ~p\n", [CP])
	    )
	;
	    write('Cannot resolve contraints')
	),
	nl.
	
try_more :-
	write('Try more (y/n)? '),
	get_line(Line),
	\+ Line = [0'y|_].

reset :-
	clean_symtab,
	clean_scope,
	retractall(current_modules(_)),
	cmd_options('-l', [LibDir]),
	atom_concat([LibDir, '/', 'Core'], CorePath),
	use_module(CorePath, [tonf/2, tohnf/2]),
	shell_load_module('Prelude'),
	open_scope(parser('Shell')).

do_shell_error(E) :-
	write(E),
	nl,
	fail.

check_constraints([], _) :- !.
check_constraints([binding(_Name, X)|XS], Prop) :-
	(
	    var(X) ->
	    reify_clp_var(X, P),
	    !,
	    Prop = P
	;
	    true
	),
	check_constraints(XS, Prop). 

%merge_reification(Set1, Set2, MergedSet).

% Debemos de crear una nueva lista con las constraints, de tal forma que
% MergedSet tenga las posibles unificaciones de los elementos de cada
% lista más los que no concuerdan con ninguno.

%% Dealing with variable naming.

%% Pretty print the bindings.
%% FIXME: We should obtain the case x = y.
write_bindings(B) :-
	write('{'),
	write_bindings2(B),
	write('}').

write_bindings2([binding(Name, X) | R]) :-
	(
	    var(X) ->
		(
		    frozen(X, _Goal) ->
		    format("Computation frozen on ~p", [Name])
		;
		    write(Name),
		    write(' is free')
		)
	;
	    write(Name),
	    write(' = '),
	    ( %% Attributed variables can fail here
		prettyvars(X) ->
		write(X)
	    ;
		write(X)
	    )
	),
	(
	    R = [] ->
	    true
	;
	    write(', '),
	    write_bindings2(R)
	).
write_bindings2([]).

%% Filter bindings obtains the original free vars present in the 
%% curry expression.
filter_bindings([], _, []).
filter_bindings([X|XS], Dicc, [binding(Name, X)|BXS]) :-
	bind_member(Name-X, Dicc),
	!,
	filter_bindings(XS, Dicc, BXS).
filter_bindings([_|L], Dicc, BL) :-
	filter_bindings(L, Dicc, BL).
	
bind_member(Name-X, [Name-BX|_]) :-
	X == BX,
	!.
bind_member(B, [_|Rest]) :-
	!,
	\+ var(Rest),
	bind_member(B, Rest).

%% Unbind expressions
unbind(Expr, Dicc, UExpr) :-
	var(Expr),
	!,
	(
	    bind_member(UExpr-Expr, Dicc)
	;
	    Expr = UExpr
	).
unbind(Expr, _Dicc, Expr) :-
	functor(Expr, _F, 0),
	!.
unbind(Expr, Dicc, UExpr) :-
	Expr =.. EList,
	!,
	unbind_list(EList, Dicc, UList),
	UExpr =.. UList.

unbind_list([], _, []).
unbind_list([L|LS], D, [R|RS]) :-
	unbind(L, D, R),
	!,
	unbind_list(LS, D, RS).

%% Modules account.

compile_curry(Module) :-
	(
	    file_name_extension(File, Module, '.curry'),
	    file_exists(File) ->
	    cmd_options('-l', [LibDir]),
	    atom_concat(['./Sloth-bin ', '-l ', LibDir, ' ', File, '> /dev/null'], CName),
	    system(CName)
	;
	    true
	).

:- data current_modules/1.

%% FIXME: Hack hack
% shell_load_module('Prelude') :-
% 	load_module(module('Prelude', [], noimport)),
% 	assert(current_modules('Prelude')),
% 	!,
% 	update_prompt.

shell_load_module(Module) :-
	(
	    \+ current_modules(Module) ->
	    catch((
		      compile_curry(Module),
		      load_module(module(Module, [], noimport))
		  ), E, do_shell_error(E)),		      
	    assert(current_modules(Module)),
	    update_prompt
	;
	    write('Module already loaded, unload first or use reload.'),
	    nl
	).

shell_unload_module(Module) :-
	(
	    current_modules(Module) ->
	    retract(current_modules(Module)),
	    unload_module(Module),
	    update_prompt
	;
	    write('Module not loaded')
	).

%% FIXME: This is not rigth.
update_prompt :-
	findall(M, current_modules(M), ML),
	build_prompt(ML, Pr),
	prompt(_, Pr).
	
build_prompt(ML, Pr) :-
	interleave_spaces(ML, ML1),
	atom_concat(ML1, ML2),
	atom_concat([ML2, '> '], Pr).

interleave_spaces([], []).
interleave_spaces([M|MS], [M,' '|MSS]) :-
	interleave_spaces(MS, MSS).

%%
is_command([0':|Command], Comm) :-
	command(Command, Comm).
	
command("quit", quit).
command(Comm, load(Module)) :-
	append("load ", Module, Comm).
command(Comm, unload(Module)) :-
	append("unload", Module, Comm).
command(Comm, reload(Module)) :-
	append("reload ", Module, Comm).
command(Comm, type(Object)) :-
	append("type ", Object, Comm).
	
do_command(quit).
do_command(load(Module)) :-
	atom_codes(AModule, Module),
	shell_load_module(AModule).

do_command(unload(Module)) :-
	atom_codes(AModule, Module),
	shell_unload_module(AModule).

do_command(reload(Module)) :-
	atom_codes(AModule, Module),
	shell_unload_module(AModule),
	shell_load_module(AModule).

do_command(type(Object)) :-
	(
	    catch(tipar(Object, Type), _E, fail) ->
%	    prolog_to_type(Type, TypeP),
	    prettyvars(Type),
	    write(Type),
	    nl
	;
	    write('Cannot type the expression, sorry'),
	    nl
	).
%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

