:- module(typer, [check_infer_types/0, tipar/2, start/0, tipo/2]).


:- use_module(symtab, [get_rules/2, set_object_property/3,
                       get_object_property/3,
                       replace_object_property/3]).  

:- use_module(curryast, [split_rule/3, get_params/2, get_function/2]).
                       % solo para pruebas

:- use_module(parser, [parse/1]).


%% Operators includes.
:- include(type_ops).
:- include(ast_ops).

% librerias que se usan para este módulo
:- use_module(library(pretty_print), [pretty_print/2]).
:- use_module(library(write),        [write/1] ).

:- use_module(library(lists),        [nonsingle/1, append/3, delete/3, nocontainsx/2, difference/3]).
:- use_module(library(terms_vars)).

%% Pruebas
:- use_module(parser,     [parse_expr/2]).
:- use_module(trans_lazy, [tr_expr/2] ).
:- use_module(curryast,   [map_tree/3] ).
:- use_module(unref,      [free_vars/3] ).
:- use_module(localdefs,  [proccess_ldef_global/2, proccess_ldef_pattern/2]).
:- use_module(modules,    [load_module/1]).
:- use_module(unref,      [free_all_type_vars/0]).

% definimos los predicados dinámicos tipot y reglat
:- dynamic tipot/2, reglat/3.

/*   *********************************************************************
	Funcionamiento del Tipador
        --------------------------

FIXME: Debido a cleanups en el typer esto ya no se corresponde con la realidad.

1. Obtiene todas los tipos predefinidos de la TS, y los almacena en el predicado dinamico tipot
2. Obtiene todas las reglas de las distintas funciones de la TS y los almacena en reglat
3. A partir de las funciones, obtiene las componentes conexas, es decir, las reglas que deben
	ser tipadas de forma conjunta. Por ejemplo, si dos funciones se llaman mutuamente, pertenecen
        a la misma componente conexa.
4. Una vez obtenidas las componentes conexas, se establece el orden en el que tienen que ser tipadas
	dependiendo de que funciones llaman a otras. (primero las que son llamadas).
5. Una vez obtenido el orden de evaluacion de las funciones (con ordeneval(X)) se procede a tipar
	cada componente conexa.
6. Para tipar una componente conexa, primero se obtiene una expresion que pueda ser tipable, a
	partir de las reglas de las funciones que intervienen. (ver Reglas de Inferencia del predicado tipo).
7. Una vez obtenidas las expresiones, se utiliza el predicado tipo para obener el tipo. 
8. Una vez obtenido el tipo, se almacena en la TS con set_object_property(funcion,type,tipo).
9. El manejador de errores guardan el mensaje de error en la tabla como si fuese un tipo, y ademas
	escribe un mensaje de error.
10. Antes de almacenar el tipo en la TS, el tipador comprueba si tenia un tipo definido. De ser asi
	hace las comprobaciones oportunas para ver que tipo tiene que guardar, si el inferido o el definido.



       Clases
       ------

1. Si una funcion llama a un operando que pertenece a alguna clase, esta se propaga
	a las variables que lo utilicen.
2. Esto se consigue utilizando etiquetas.
3. Si la variable con etiqueta se unifica con un Atomo, se elimina la etiqueta y se unifica.
4. Si la variable con etiqueta se unifica con otra variable, si esta no tiene etiqueta, se
unifica con etiqueta, y si la tiene, se comprueba si la clase de la etiqueta es compatible.
5. las operaciones con clases seran de la forma (== :: clase(eq,X) => X -> X -> Bool)


*/



/*
 Cuestiones por resolver:


2. Existen varios tipos escritos de dos maneras: p.e. 'Nat' y qid('Main','Nat')
   o 'Bool' y qid('Prelude','Bool'),
   y sobre todo [A] y qid('Prelude',[A])  hay que unificar criterios.(¿Resuelto?)



6. en la TS habría que almacenar los tipos definidos como typedef o deftype, pero no
   como type, porque el tipador se confunde con tipos válidos.
7. si el tipo definido no tiene eq(X)=> y el inferido si, ¿debe dar error? Yo creo que no.

NOTAS:

- Los Warnings los trato como errores.
- los manejadores de errores, guardan el mensaje de error en la tabla como si fuera un
tipo, y además emiten un mensaje de error.
- las clases se obtienen únicamente de los tipos predefinidos. Si una funcion predefinida
tiene una clase, esta se propaga a todas las funciones que la usen.


*/
 


%       Extraemos los constructores, data y reglas de la tabla
%       ------------------------------------------------------


% listatipos saca los tipos de los constructores en una lista
listatipos(Rs) :-
	findall((X, Y), get_object_property(X, type, Y), Rs).



% instipo guarda los tipos de los constructores a partir de listatipos
instipo([]).
instipo([(A, B)|Xs]) :-
	assert(tipot(A, B)), 
	instipo(Xs).


% Tablareglas obtiene la lista de reglas de la tabla de Simbolos (TS)
tablareglas(R) :-
	findall((A, X), get_rules(A, X), L), 
	listaReglas(L, R).



% convierte la lista de reglas obtenida por tablareglas(L) en una
% unica lista de arboles sintacticos.
listaReglas([],[]).
listaReglas([X|Xs],Rs) :-
	lreglas2(X,A1), 
	listaReglas(Xs,A2), 
	append(A1,A2,Rs).

lreglas2((_, [builtin(_)]), []) :- !.
% Temporal (quitar si se soluciona el unref)
lreglas2((_A, X), []) :- 
	esunref1(X).
lreglas2((_A, B), B).


% para quitar el unref 
esunref1(X) :-
	esunref(X,N),
	N>0.

esunref(unref(_), 1).
esunref(A:=B, X+Y) :-
	esunref(A,X),
	esunref(B,Y).
esunref(A@B, X+Y) :-
	esunref(A,X),
	esunref(B,Y).
esunref(const(_A,_B),0).
esunref(qid(_A,_B),0).
esunref(param(_),0).
esunref(A,X) :-
	list(A),
	esunrefl(A,X).

esunrefl([], 0).
esunrefl([A|As], X+Y) :-
	esunref(A,X),
	esunrefl(As,Y).


% insregla transforma la lista de reglas obtenidas con tablareglas
% al formato adecuado para ser tipadas, y las guarda.
insregla([]).
insregla([X|Xs]) :-
	split_rule(X, A, B), 
        get_params(A, R1), 
	get_function(A, A1), % XXX
        l2var(regla(A1, R1, B), Reg, Lvar), %% XXX
        annadevars(Lvar, Lvar2, _W),        %% XXX
        sust9(Lvar2, Reg, Reg2),            %% XXX
        Reg2=regla(RR1, RR2, RR3),
        assert(reglat(RR1, RR2, RR3)),
    	insregla(Xs).

% por si la regla no respeta el formato.
insregla([X|Xs]) :-
	X\=(_A:=_B), insregla(Xs). 


% este es el predicado principal, que obtiene los tipos predefinidos, los constructores
% y las reglas de la tabla, y las guarda en predicados dinamicos tipot y reglat.
% a partir de estos obtiene los tipos de las reglas ( si es posible )
% y los almacena en la tabla.

check_infer_types :-
	listatipos(Rs), % XXX
	instipo(Rs), % obtiene los tipos predefinidos XXX
        tablareglas(L), % XXX
	insregla(L), % obtiene las reglas de la TS XXX
        todaslasreglas(X), % Calcula los tipos de las reglas
        gtipotabla(X). % Guarda las reglas en la TS         XXX


/*
l_tipos2(X):-   % para CLASES
             update(==,type, eq(_G)=>_G --> _G --> 'Bool'),
             listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X), % Calcula los tipos de las reglas CON CLASES
             verErrores(Errores). % esta funcion ya no existe (poner gtipotabla(X))
  
*/


%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%  PREDICADO TIPO
%%%%%%%%%%%%%%%%%%%%%%%%%%

tipot(A, T) :-
	get_object_property(A, type, T).


%                     Tipos básicos predefinidos
%                     -------------------------

% asi deberia ser
tipo(const(Type, _Value), Type) :- !.
% tipo(const(int,_), 'Int') :- !.

tipo([X|[]], list(Y)) :-
	(tipo(X, Y) ; throw(error_typer('Different types in list'))). 
tipo([X|Xs], list(Y)) :-
	tiplista(X, Y),
	Xs\=[],
	tipo(Xs, list(Y)).


% tipo tupla
tipo((X, Y), (A, B)) :-
	tipo(X, A), 
	tipo(Y, B).


%                 tipos derivados de la tabla de simbolos   
%                 ---------------------------------------

% ¿Estos tipos todavia valen?
tipo(num(N), X) :-
	(
	    integer(N), 
	    X='Int', 
	    ! 
	;
	    X='Float',
	    !
	).

tipo(char(_), 'Char').
tipo(str(_), ['Char']).

% por si acaso
tipo(X, 'Float') :-
	float(X), !.
tipo(X, 'Int') :-
	int(X), !.


%                             Reglas de inferencia
%                             --------------------


% los tipos sacados de la tabla de simbolos
tipo(A, T) :-
	tipot(A, T), !.

% Variables
tipo(var(X), X).


% abstracciones
tipo(lambda(P, E), (T1-->T2)) :-
	mis1(P, T1), 
	mis1(E, T2).



tipo(case([]), _X).
tipo(case([L1|Ls]), (T1-->T2)) :-
	mis2(L1, (T1-->T2)), % o debe ser mis1 ??
	tipo(case(Ls), (T1-->T2)).

tipo(F@[], Tipo) :-
	!,
	tipo(F, Tipo).

tipo((F @ X), T2) :-
	(
	    tipo(F, (T1 --> T2)), 
	    mis1(X, T1), 
	    ! 
						% MODIF para MiSMATCH
	; 
	    tipo(F, (=>(clase(Clase,A), (T1 --> T2)))),
	    attach_attribute(A, (Clase, A, _Att)),
	    mis1(X, T1)
	) .

/*
% Let (¿Esto ya no vale?)
tipo(le(X, E1, E2), T) :-
	tipo(E1, T1), 
	tipo(X, T1), 
	tipo(E2, T).
*/

% Fixed Point
tipo(fix(X, E), T) :-
	tipo(X, T), 
	tipo(E, T).


tiplista(X, Y) :-
	(tipo(X, Y) ; throw(error_typer('Different types in list'))).    


% para comprobar errores type Mismatch
mis1(P, T1) :-
	tipo(P, T1) 
 ; 
	throw(error_typer('Type Mismatch')).


mis2(P, T1) :-
	tipo(P, T1) 
 ; 
	throw(error_typer('Rules have diferrent types')).




% ***********************  CLASES!!!!

% Comparar_clases

clases_predefinidas(X) :-
	X=[(eq, Eq), (show, Show), (read, []), (bounded, []),
        (monad, Monad), (functor, []), (enum, Enum)],
        Eq=[(ord, Ord), (num, Num)],
        Show=[(num, Num)],
        Ord=[(real, Real)],
        Num=[(real, Real), (fractional, Fractional)],
        Enum=[(integral, [])],
        Real=[(integral, []), (realFrac, RealFrac)],
        Fractional=[(realFrac, RealFrac), (floating, Floating)],
        RealFrac=[(realFloat, [])],
        Floating=[(realFloat, [])],
        Monad=[(monadPlus, [])].


comparar_clases(X, X, X) :- !.                                      
comparar_clases(X, Y, Y) :-
	subclases(X, L), 
	esta_clase2(Y, L), !. 
comparar_clases(X, Y, X) :-
	subclases(Y, L), 
	esta_clase2(X, L), !.
comparar_clases(X, Y, Z) :-
	comparten_clase(X, Y, Z), !.


/*
% esto se usa?
esta_clase(A, []) :- 
	fail.
esta_clase(A, [(A, Lb)|S]).
esta_clase(A, [(B, Lb)|S]) :-
	A\=B, 
	esta_clase(A, S). 
*/


esta_clase2(_A, []) :-
	fail, !.
esta_clase2(A, [(A, _Lb)|_S]) :-	!.
esta_clase2(A, [(B, Lb)|S]) :-
	A\=B, 
	(esta_clase2(A, S), !) 
 ;      (esta_clase2(A, Lb), !).

comparten_clase(X, Y, Z) :-
	subclases(X, Lx), 
	subclases(Y, Ly), 
	comparten2(Lx, Ly, Z).

comparten2([(Z, _La)|_S], Ly, Z) :-
	esta_clase2(Z, Ly), !.
comparten2([(_A,La)|S],Ly,Z) :- 
	(comparten2(S, Ly, Z),!) ;
        (comparten2(La, Ly, Z),!) . 


subclases(A, L) :-
	clases_predefinidas(C), 
	esta_clase2(A, C),
        lista_subclases(A, L, C).

lista_subclases(A, L, [(A, L)|_S]) :- !.
lista_subclases(A, L, [(B, Lb)|_S]) :-
	A\=B, esta_clase2(A, Lb),
        lista_subclases(A, L, Lb), !. 
lista_subclases(A, L, [(B, _Lb)|S]) :-
	A\=B, esta_clase2(A, S),
        lista_subclases(A,L,S), !.
                            

% Multifiles verify_attribute/2 y combine_attribute/2

:- multifile verify_attribute/2.
verify_attribute(A, T) :-
	A=(_Clase, Variable, _Att), 
	detach_attribute(Variable), 
	Variable=T.

:- multifile combine_attributes/2.
combine_attributes(V1, V2) :-
	V1=(Clase1, Variable1, Att1),
	V2=(Clase2, Variable2, _Att2),
	( comparar_clases(Clase1, Clase2, Clase3)
 ;
	throw(error_typer('Class Mismatch'))),    
	Variable1=Variable2,
	update_attribute(Variable1,(Clase3, Variable1, Att1)).

% para pruebas
ee(X,Y) :-
	varset(X, Y).


% FIN CLASES


% tipador de reglas

%        convierte las reglas en una sola expresión con lambdas.
%        ------------------------------------------------------

elambdaux(_N, S, [], S).
elambdaux(N, lambda(X, E), [X|Xs], S) :-
	elambdaux(N, E, Xs, S).

elambda(N, E) :- 
	regla(N, L, S), elambdaux(N, E, L, S).
 

% modificamos ecase para que sólo ponga case si hay varias reglas
ecase(N, case(L)) :-
	setof(X, elambda(N, X), L), 
        nonsingle(L).     %	\+ hayuna(L).
ecase(N, X) :-
	setof(X, elambda(N, X), L), 
	\+ nonsingle(L),       % hayuna(L), 
	L=[X|[]].

lfix([],[]).
lfix([X|Xs], [Y|R]) :- 
	ecase(X, Y), 
	lfix(Xs, R).


% saca la expresión con fix de una componente conexa
% y  sustituye los nombre de reglas por var(X) en los fix,
%  y lo convierte en tuplas.
ffix2(L, L1) :-
	\+ nonsingle(L),  % hayuna(L), 
	L=[A], 
	\+esrecur(A), 
	ecase(A, L1), 
	!.
ffix2(L, R) :-
	\+ nonsingle(L),  % hayuna(L),
	L=[A], 
	esrecur(A), 
	ecase(A,B), 
	sust(A, fix(A, B), R), 
	!.
ffix2(L, fix(X, Y)) :-
	nonsingle(L), % \+hayuna(L), 
	lfix(L,L1),
        reemplazo(L, fix(L, L1), R), 
	R=fix(A,B), 
	listatupla(A, X), 
	listatupla(B, Y),
	!.

tipear([],[]).
tipear([X|Xs], [W, L1]) :- 
	ffix2(X, E), 
	tipoerror(E, T), 
	gua(X, T, W), 
	tipear(Xs, L1).

/*
% si la expresión es no tipable, se le pone el tipo No Tipable.
tipoerror(E, T) :-
	(tipo(E, T), !) 
 ;      (T='No Tipable',!).

*/

tipoerror(E, T) :-
	catch(tipoerror2(E,T), error_typer(Mensaje), T=error(Mensaje)).  % Catch provisional

% PARA CLASES

tipoerror2(E, T2) :- 
	(tipo(E, T), pclas(T,T2), ! ).
% ;      (T2='No Tipable',!).


% obtiene la lista de variables de un tipo T y añade las clases, si las
% variables tenian atributos asociados.
pclas(T, T2) :- 
	varset(T, L), 
	ponerclases(T, L, T2).


% a partir de un tipo T y de una lista de variables [A|As]
% añade las clases Eq(X)=> X etc.. al tipo original en T2
ponerclases(T, [], T).
ponerclases(T, [A|As], T2) :- 
	(get_attribute(A, (A1, A2, _A3)),
	 detach_attribute(A),
        ponerclases((clase(A1,A2)=>T), As, T2), !) ;
        (ponerclases(T, As, T2), !).



/*
% a partir de un tipo T y de una lista de variables [A|As]
% añade las clases Eq(X)=> X etc.. al tipo original en T2
ponerclases(T, [], T).
ponerclases(T, [A|As], T2) :- 
	(get_attribute(A, (A1, A2)),
        ponerclases((A1=>T), As, T2), !) ;
        (ponerclases(T, As, T2), !).
*/
% FIN CLASES


% guarda el tipo obtenido por si es necesario para tipear otra regla
gua([A|B], (X, Y), [(A, X), Rs]) :- 
	B\=[], 
	assert(tipot(A, X)), 
	gua(B, Y, Rs).
gua([A|[]], T,(A, T)) :- 
	assert(tipot(A, T)), !.


% obtiene el tipo de todas las reglas y lo guarda en la lista T
todaslasreglas(T) :- 
	ordeneval(L), 
	tipear(L,L1),
	li(L1,T).


% para guardar la lista de tipos de reglas en la tabla de simbolos      
gtipotabla([]).
gtipotabla([X|Xs]) :- 
	X=(A, B),
	% Comprobar si hay tipo definido antes de guardar el tipo inferido !!!!!!!
	write(A), write('  -  '), pretty_print(B,[]), write('\n'),
        (B=error(_)   ;  check_type(X)),
	gtipotabla(Xs).

% obtiene el tipo de una regla (se usa para ejemplos)
% tiporegla(X,T):- cconexa(X,L), ffix2(L,E), tipo(E,T), pretty_print(T,[]).
tiporegla(X, T) :- 
	cconexa(X, L), 
	ffix2(L, E), 
	tipo(E, T).


%  predicados para convertir var(x) en var(X) en las reglas         
% ---------------------------------------------------------

  
% unir une dos listas sin elementos repetidos

unir([], [], []) :- !.
unir([], Y, Y) :- !.
unir(X, [], X) :- !.
unir([X|Xs], Y, [X|W]) :- 
	nocontainsx(Y, X),    %noesta(X, Y), 
	unir(Xs, Y, W).
unir([X|Xs], Y, W) :- 
        \+ nocontainsx(Y,X),   %	\+ noesta(X, Y), 
	unir(Xs, Y, W).
   
% para usar despues de sacar la lista de reglas
annadevars([], [], var(_O)).
annadevars([var(N)|Rs], [W|Zs], (Y, L)) :- 
	copy_term(var(_X),Y),
        W=(var(N), Y),                                                                 
	annadevars(Rs, Zs, L). 


% l2var coge una regla y la transforma en regla valida
% y saca la lista de variables que tiene.

%% FIXME: Este lo matamos.

l2var(regla(A, B, C), regla(X, Y, Z), L) :- 
	l2var(A, X, L1),
        l2var(B, Y, L2), 
	l2var(C, Z, L3), 
	unir(L1, L2, L4),
        unir(L4, L3, L).
% l2var(id(A), A, []).
% l2var(var(A), var(A), [var(A)]).
l2var(qid(A, B), qid(A, B), []).            % EMILIO
l2var(param(A), var(A), [var(A)]).         % EMILIO
l2var(const(A, B), const(A, B), []).         % EMILIO
l2var((A@B), (X@Y), L) :- 
	\+list(B) , 
        l2var(A, X, L1), 
	l2var(B, Y, L2), 
	unir(L1, L2, L).

%% ERROR. El problema es que, al llamar a l2var despues de l2varaux
%% vuelve a ejecutarse l2varaux y el tipo a@[b] no aparece nunca.
%% aparece a@b

l2var((A@B), X, L) :- 
	list(B),
	l2varaux((A@B), C), 
	l2var2(C, X, L).

     

l2var((A, B), (X, Y), L) :- 
	l2var(A, X, L1),
	l2var(B, Y, L2),
	unir(L1, L2, L).
l2var(A, X, L) :- 
	list(A), 
	ll2var(A, X, L), !.
l2var(num(X), num(X), []).
l2var(char(X), char(X), []).
l2var(str(X), str(X), []).
 
% para usar l2var con listas
ll2var([], [], []) :- !.
ll2var([A|As], [X|Xs], L) :- 
	l2var(A, X, L1), 
	ll2var(As, Xs, L2), 
	unir(L1, L2, L).

% para convertir las expresiones A@[ B1 B2 ] en expresiones tipables
% como ((A@B1)@B2)
 
l2varaux((A@[]), A).
l2varaux((A@[B|Bs]), C) :- 
	B\=(_M@_N),
	X=(A@B),
	l2varaux((X@Bs), C).
l2varaux((A@[B|Bs]), C) :- 
	B=(_M@_N), 
	l2varaux(B, B1), 
	X=(A@B1), 
	l2varaux((X@Bs), C).





%%%%%%%%%% AUXILIAR DE L2VAR %%%%%%%%%%%% L2VAR2
%%%%%%
%%%  Se usa para dar una segunda pasada después de examinar un formato A@B
%%% sin que vuelva a examinar A@B.


% l2var2 coge una regla y la transforma en regla valida
% y saca la lista de variables que tiene.

l2var2(regla(A, B, C), regla(X, Y, Z), L) :- 
	l2var2(A, X, L1), 
	l2var2(B, Y, L2), 
	l2var2(C, Z, L3),
        unir(L1, L2, L4),
	unir(L4, L3, L).
% l2var2(id(A), A, []).
% l2var2(var(A), var(A), [var(A)]).
l2var2(qid(A, B), qid(A ,B), []).             % EMILIO
l2var2(param(A), var(A), [var(A)]).           % EMILIO
l2var2(const(A ,B), const(A, B), []).         % EMILIO
l2var2(unref(A), unref(A), []).               % EMILIO
l2var2((A@B), (X@Y), L) :-  
	l2var2(A, X, L1), 
	l2var2(B, Y, L2), 
	unir(L1, L2, L).
l2var2((A, B), (X, Y), L) :- 
	l2var2(A, X, L1),
	l2var2(B, Y, L2),
	unir(L1, L2, L).
l2var2(A, X, L) :- 
	list(A), 
	ll2var2(A, X, L), !.
l2var2(num(X), num(X), []).
l2var2(char(X), char(X), []).
l2var2(str(X), str(X), []).
   
% para usar l2var2 con listas
ll2var2([], [], []) :- !.
ll2var2([A|As], [X|Xs], L) :- 
	l2var2(A, X, L1), 
	ll2var2(As, Xs, L2), 
	unir(L1, L2, L).

%%%%% FIN AUXILIAR
%%%%%%%%%%%%%%%%%%%


% sustituye una lista de variables var(x) por sus var(X) correpondientes
% en la expresión E
sust9([], E, E).
sust9([(A, B)|Xs], E, E2) :- 
	sust5(A, E, B, E1), 
	sust9(Xs, E1, E2).
  
%  Sustituye N por S, se usa para sustituir var(x) por var(X) 
% en las reglas obtenidas por la TS, para que sean tipables.
sust5(N, N1, S, S) :- 
	N==N1.
sust5(N, regla(A, B, C), S, regla(X, Y, Z)) :- 
	sust5(N, A, S, X), 
	sust5(N, B, S, Y),
        sust5(N, C, S, Z).
sust5(N, (A@B), S, (J@K)) :- 
	sust5(N, A, S, J),
	sust5(N, B, S, K).
sust5(N, (A, B), S, (J, K)) :- 
	sust5(N, A, S, J),
	sust5(N, B, S, K).
sust5(N, B, _S, B) :- 
	atom(B), 
	\+ list(B), 
	N\=B.
sust5(_N, num(X), _S, num(X)).
sust5(_N, const(A, B), _S, const(A, B)).           % EMILIO
sust5(_N, qid(A, B), _S, qid(A, B)).               % EMILIO
sust5(_N, unref(A), _S, unref(A)).                 % EMILIO
sust5(_N, char(X), _S, char(X)).
sust5(_N, str(X), _S, str(X)).
sust5(_N, B, _S, B) :- 
	number(B).
sust5(N, var(A), _, var(A)) :- 
	\+(N==var(A)).
sust5(N, L, S, L1) :- 
	list(L), 
	sustl5(N, L, S, L1).

% lo mismo que sust5, por si la expresión E es una lista
sustl5(_N,[],_S,[]).
sustl5(N, [X|Xs], S, [A|As]) :- 
	sust5(N, X, S, A), 
	sustl5(N, Xs, S, As).
   
%               Calculo de componentes conexas para fix
%               ---------------------------------------

% predicado para ver si un nombre aparece en una expresion
esta(N, N).
esta(N, lambda(X, Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.
esta(N, (X@Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.
esta(N, le(X, Y, Z)) :- 
	esta(N, X), !
 ;      esta(N, Y), ! 
 ;      esta(N, Z), !.
esta(N, (X, Y)) :- 
	esta(N, X), ! 
 ;      esta(N, Y), !.

% predicado para sustituir el nombre de la regla por una var(X)
% y usarlo en el tipado de fix (punto fijo)
sust(N, N, var(_X)).
sust(N, qid(X, Y), qid(X, Y)) :-
	N\=qid(X,Y).
sust(N, lambda(A, B), lambda(J, K)) :- 
	sust(N, A, J), 
	sust(N, B, K).
sust(N, (A@B), (J@K)) :- 
	sust(N, A, J),
	sust(N, B, K).

/*
sust(N, le(A, B, C), le(J, K, L)) :- 
	sust(N, A, J),
	sust(N, B, K),
	sust(N, C, L).

*/
sust(N, (A, B), (J, K)) :- 
	sust(N, A, J),
	sust(N, B, K).
sust(N, B, B) :- 
	atom(B), 
	\+ list(B), 
	N\=B.
sust(_N, B, B) :- 
	number(B).
sust(_N, var(A), var(A)).
sust(_N, const(A,B), const(A,B)).        % EMILIO

% estas siguen haciendo falta?
sust(_N, num(X), num(X)).
sust(_N, id(X), id(X)).
sust(_N, char(X), char(X)).
sust(_N, str(X), str(X)).

sust(N, L, L1) :- 
	list(L), 
	sustl(N, L, L1).
sust(N, case(X), case(Y)) :- 
	sust(N, X, Y).
sust(N, fix(X, Y), fix(A, B)) :- 
	sust(N, X, A),
	sust(N, Y, B).

% como sust, pero para listas
sustl(_N, [], []).
sustl(N, [X|Xs], [A|As]) :- 
	sust(N, X, A), 
	sustl(N, Xs, As).
 
% reemplaza todas las variables en un fix([],[]) (A debe ser fix)    
reemplazo([], A, A).
reemplazo([X|Xs], A, Z) :- 
	sust(X, A, B), 
	reemplazo(Xs, B, Z).

reemplazo2(fix(A, B), C) :- 
	reemplazo(A, fix(A, B), C).
   
% predicado para obtener la expresión final tipable
efinal(N, fix(N, case(L))) :- 
	elambda(N, Y), 
	esta(N, Y),
        setof(X, elambda(N, X), L), !.
efinal(N, case(L)) :- 
	setof(X, elambda(N, X), L).

% para ver si una función se llama a si misma
esrecur(X) :- 
	regla(X, _, Y), 
	esta(X, Y).

% para calcular funcions que llaman a otras funciones
% apunta(N,E) <-- N apunta a E.
apunta(N, E) :- 
	regla(N, _, Y), 
	esregla(E), 
	N\=E,  
	esta(E, Y).

% apunta, si el segundo término es una lista.
% para ver si una componente conexa apunta a una función).
apunta2(N, [A|[]]) :- 
	apunta(N, A).
apunta2(N, [X|Xs]) :- 
	apunta(N, X), 
	Xs\=[], ! 
 ;      apunta2(N, Xs).

% apunta, si el primer término es una lista.
% (para ver si una funcion apunta a una componente conexa).
apunta3([X|[]], Y):- 
	apunta(X, Y).
apunta3([X|Xs], Y) :- 
	apunta(X, Y), 
	Xs\=[], ! 
 ;      apunta3(Xs, Y).

% apunta si los dos términos son listas (para ver si dos componentes
% conexas se apuntan)
apunta1([X|[]], [Y|Ys]) :- 
	apunta2(X, [Y|Ys]).
apunta1([X|Xs], [Y|Ys]) :- 
	apunta2(X, [Y|Ys]), 
	Xs\=[], ! 
 ;      apunta1(Xs, [Y|Ys]).



%                    Predicados para Listas
% algunas se podrían sustituir por las que están en la libreria list
% (programadas en un momento en el que no se disponia de la documentación
% de la libreria list)
%  -----------------------------------------------

% list_concat().
% predicado para convertir una lista de listas en una lista
li([X|Xs], [X|Ys]) :- 
	\+ list(X), 
	li(Xs, Ys).
li([X|Xs], R) :- 
	list(X), 
	li(X, Y), 
	li(Xs, Ys), 
	append(Y, Ys, R).
li([], []).
  
% convierte una lista en una tupla
listatupla([A|As], (A, Bs)) :- 
	listatupla(As, Bs).
listatupla([X|[]], X).

/*

% indica si un elemento NO ESTA en una lista. (es como nocontainsx(L,X) ).
noesta(A, []).
noesta(A, [X|Xs]) :- 
	A\=X, 
	noesta(A, Xs).

% elimina un elemento de una lista (es como delete(L1,E,L2) )
quitar(A, [], []).
quitar(A, [A|Xs], Ys) :- 
	quitar(A, Xs, Ys).
quitar(A, [X|Xs], [X|Ys]) :- 
	A\=X, 
	quitar(A, Xs, Ys).

% elimina los elementos de una lista de otra lista  (es como difference(Linicial,Lquitar,Lfinal)
quitarl([], Xs, Xs).
quitarl([A|As], Xs, Zs) :- 
	quitar(A, Xs, Ys), 
	quitarl(As, Ys, Zs).

*/

% ordena una lista de funciones que se apuntan unas a otras
ordenar([], Xs, Xs).
ordenar([A|As], Bs, Zs) :- 
	inserta(A, Bs, Cs), 
	ordenar(As, Cs, Zs).

% inserta una funcion dependiendo de si apunta o no a las demás
inserta(A, [], [A]).
inserta(A, [B|Bs], [A|Cs]) :- 
	apunta1(B, A), 
	Cs=[B|Bs].
inserta(A, [B|Bs], [B|Cs]) :- 
	\+ apunta1(B, A), 
	inserta(A, Bs, Cs).

% devuelve una lista sin elementos repetidos 
norep([],[]).
norep([X|[]], [X|[]]).
norep([X|Xs], [X|Zs]) :- 
	Xs\=[], 
	delete(Xs, X, Ys),    % quitar(X, Xs, Ys), 
	norep(Ys, Zs).

% devuelve la lista de reglas en R
lreglas(R) :- 
	findall(X, regla(X, _, _), L), 
	norep(L, R).


/*
% indica si la lista sólo tiene un elemento
hayuna([X|[]]).
*/


%                Predicados para Reglas
%                ----------------------
esregla(X) :- 
	lreglas(L), 
	pertenece(X, L).
 
pertenece(X, [X|[]]).
pertenece(A, [X|Xs]) :- 
	A=X ; pertenece(A, Xs).

%funcion para calcular el camino con una lista de grafos
cam(N, Y) :- 
	camaux(N, Y, []).
 
camaux(X, X, _L).
camaux(X, Y, L) :- 
	apunta(X, Z), 
	nocontainsx(L,Z),   % noesta(Z, L), 
	camaux(Z, Y, [Z|L]).

% Cálculo de la componente conexa usando cam
cconexa2(X, Y):- 
	cam(X, Y), 
	cam(Y, X).

cconexa(X, L) :- 
	setof(Y, cconexa2(X, Y), L).

% calculo de la lista de componentes conexas a partir de la lista de reglas
lconexa(L) :- 
	lreglas(R), 
	lconexa2(R, L).

lconexa2([], []).
lconexa2([X|Xs], [Y|Ys]) :- 
	cconexa(X, Y), 
	difference([X|Xs], Y, Zs),  % quitarl(Y, [X|Xs], Zs),
        lconexa2(Zs, Ys).

% devuelve la lista de componentes conexas ordenadas para su evaluacion
ordeneval(L) :- 
	lconexa(R), 
	ordenar(R, [], L).  



%                             Reglas de la TS
%                            ------------------
regla(A,B,C) :- 
	reglat(A,B,C).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%  TRATAMIENTO DE ERRORES

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% función para comprobar si el tipo inferido
%% concuerda con el especificado.

%% comprobar_tipos

% si el tipo inferido y el definido son iguales, todo queda igual
% si el tipo inferido es más restrictivo, da error
% si el tipo definido es el más restrictivo, se da un warning 
%  y se usa este tipo.



check_type((Regla,InfrType)) :-
	( get_object_property(Regla,type,DefType),	 
	 catch(comprobar_tipo(Regla,InfrType,DefType),error_typer(Mensaje), 
	         (write(Regla),write(' - '),write(Mensaje), write('\n')))
	)
 ;
	set_object_property(Regla,type,InfrType).

comprobar_tipo(Regla, InfrType, DefType) :- 
	( numbervars(DefType,  0, _),
	  numbervars(InfrType, 0, _),
	  InfrType==DefType
	)  
	% el tipo inferido y el definido son iguales,
        % dejamos en la tabla el tipo definido.    
 ;
	comprobar_tipo2(Regla,InfrType,DefType).
	
comprobar_tipo2(Regla,InfrType,DefType) :-
	(
	InfrType=DefType,  % unificamos los tipos
	comprobar_tipo3(Regla,DefType)
        )
 ;
	% el tipo inferido y el definido son diferentes
        %write('error1').
	throw(error_typer('Inferred type does not match the specified one')). % typerror(Regla,11)
	
comprobar_tipo3(Regla,X) :-
	( %%  lookup(Regla,defType,DefType),
	get_object_property(Regla,type,DefType),
	numbervars(DefType,0,_),
	numbervars(X,0,_),
	DefType==X,
	% El tipo inferido es más restrictivo, lo sustituimos
	% y enviamos un warning
	replace_object_property(Regla,type,X),
        %write('error warning')
        
	throw(error_typer('WARNING: Inferred type does not match the specified one'))
	)
 ;
        write('error3').
	% throw(error_typer('Inferred type does not match the specified one')) %  typerror(Regla,11) 
	
                   
/*

% Lista de errores de tipos
mensaje_error(_,Num_Error,Mensaje):-
	Num_Error=1, Mensaje='Undefined Symbol'  % no se usa
 ;          Num_Error=2, Mensaje='Different types in list'
 ;          Num_Error=3, Mensaje='Type Mismatch'
 ;          Num_Error=4, Mensaje='Undefined Constructor'  % No se usa
 ;          Num_Error=5, Mensaje='A constraint type was expected'
 ;          Num_Error=6, Mensaje='Then and Else expressions have different types' % seria type mismatch
 ;          Num_Error=7, Mensaje='If expression must be boolean'  % sería type mismatch
 ;          Num_Error=8, Mensaje='Invalid constraint in choice expression'
 ;          Num_Error=9, Mensaje='Expression in choice are not the same type'
 ;          Num_Error=10, Mensaje='Rules have different types'
 ;          Num_Error=11, Mensaje='Inferred type does not match the specified one'
 ;          Num_Error=12, Mensaje='Infinite type for function'
 
% A partir de aqui van los warnings

    ;          Num_Error=13, Mensaje='Inferred type does not match the specified one'
    ;          Num_Error>13, Mensaje=0
    ;          Num_Error<1,  Mensaje=0.


*/

start :-
	load_module(module('Prelude', [], noimport)).

tipar(Input, Tipo) :-
	catch(parse_expr(Input, Expr), E, write(E)),
	map_tree(Expr, proccess_ldef_global, Expr1),
	map_tree(Expr1, proccess_ldef_pattern, LExpr),
	map_tree(LExpr, free_vars(_Dicc), ExprFree),
	l2var(ExprFree, ExprTipos, _),
	tipo(ExprTipos, Tipo).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

